/*
 * trax.c -- Memory allocation functions
 *
 * This file contains routines to track memory allocation.
 *
 * The user may free all memory allocated so far or until the last marker.
 * Also, the user may specify allocation types and then free them selectively
 * later.
 *
 * Note:
 *		free() is never needed again by the user.  However, it may be used
 *		to minimize the maximum amount of memory used at any one moment.  It
 *		is provided for backward compatibility as well.
 *
 * Example:
 *
 *	#include "trax.h"
 *	...
 *	main()
 *	{
 *		char *p;
 *
 *		p = malloc(23);
 *		malloc(220);
 *		malloc(10);
 *
 *		mWipe();		-- deallocate all memory.
 *	}
 *
 *	Another example:
 *
 *	#define TreeNodes	1
 *	#define	Misc		2
 *
 *	main()
 *	{
 *		mMark(TreeNodes);
 *		malloc(...)
 *		malloc(...)
 *		mMark(Misc);
 *		malloc(...)
 *
 *		mWeed(TreeNodes);	-- deallocate all TreeNodes
 *		mWeed(Misc);		-- deallocate all misc memory
 *	}
 *
 * The following routines are available:
 *
 *		mMalloc()		-- Same as malloc, but track memory
 *		mCalloc()		-- Same as calloc, but track memory
 *		mrealloc()		-- Same as realloc, but track memory
 *		mFree()			-- Free the pointer passed in
 *		mMark()			-- Memory from now on is marked as type 'code'
 *		mWeed()			-- Free all memory blocks with certain type
 *		mWipe()			-- Free all memory
 *		mSetErr()		-- Specify an error reporting function
 *		mLog(f,code)	-- Print a report to stream f
 *		mHist(f)		-- Print a report of all types to stream f
 *		mMax			-- Global unsigned int that reflects max mem allocated
 *		mTotal			-- Global unsigned int that reflects total mem allocated
 *		mCur			-- Global unsigned int reflecting current mem allocated
 *
 * Address Alignment:
 *
 *		All pointers returned by mCalloc, mMalloc are sizeof(Trax)+base;
 *			where base is returned by the malloc().
 *
 * Memory corruption detection:
 *
 *		Memory corruption detection is possible when a buffer is overrun by
 *			1 byte.
 *		Magic words are 2 chars long and are accessed independently to avoid
 *			memory address alignment problems on different machines.
 *
 * Note: group array must be initialized to 0 by compiler
 *
 * The user identifies/marks memory with a code from 0..nCodes-1.  They are
 * indices into a table of pointers to linked lists.  0 is the default memory
 * code.
 *
 * Terence Parr
 * Purdue University Electrical Engineering
 * December 1989
 */

#include <stdio.h>
#include "trax.h"
#undef malloc
#undef free
#undef calloc
#undef realloc

#define bufstart(m)		((char *)(m))+sizeof(Trax)
#define magicwd1(m)		(*(bufstart(m)+m->size))
#define magicwd2(m)		(*(bufstart(m)+m->size+1))
#define badmagic(m)		( magicwd1(m) != MagicWd1 && magicwd2(m) != MagicWd2 )
#define nCodes			20

static Trax *group[nCodes];
static int Code = 0;
unsigned mMax = 0, mTotal = 0;
unsigned mCur = 0;

void defErr();
static void (*mError)() = defErr;

char *
mMalloc(n, code, file, line)
unsigned n, line, code;
char *file;
{
	register Trax *m, *h;

	if ( code >= nCodes )
	{
		(*mError)("malloc: memory code %d out of range 0..%d\n", code, nCodes-1); 
		return( NULL );
	}
	if ( code == 0 ) code = Code;
	h = group[code];
	m = (Trax *) malloc( n+sizeof(Trax)+(2*sizeof(char)) );
	if ( m == NULL )
	{
		(*mError)("malloc: can't allocate %d bytes at %s(%d)\n", n, file?file:"???", line);
		return( NULL );
	}
	m->size = n;
	m->line = line;
	m->file = file;
	m->code = code;
	m->mword = ValidBuf;
	if ( h != NULL ) h->prev = m;
	m->next = h;
	m->prev = NULL;
	group[code] = m;
	magicwd1(m) = MagicWd1;
	magicwd2(m) = MagicWd2;
	mCur += n;
	mTotal += n;
	if ( mCur > mMax ) mMax = mCur;
	return( bufstart(m) );
}

char *
mCalloc(n, s, code, file, line)
unsigned n, s, line, code;
char *file;
{
	register unsigned bytes = n*s;
	register Trax *m, *h;

	if ( code >= nCodes )
	{
		(*mError)("calloc: memory code %d out of range 0..%d\n", code, nCodes-1); 
		return( NULL );
	}
	if ( code == 0 ) code = Code;
	h = group[code];
	m = (Trax *) malloc( bytes+sizeof(Trax)+(2*sizeof(char)) );
	if ( m == NULL )
	{
		(*mError)("calloc: can't allocate %d bytes at %s(%d)\n", n, file?file:"???", line);
		return( NULL );
	}
	m->size = bytes;
	m->line = line;
	m->file = file;
	m->code = code;
	m->mword = ValidBuf;
	{	register char *p, *endp;
	
		p = bufstart(m),
		endp = p+bytes;
		do { *p++ = '\0'; } while ( p < endp );
	}
	if ( h != NULL ) h->prev = m;
	m->next = h;
	m->prev = NULL;
	group[code] = m;
	magicwd1(m) = MagicWd1;
	magicwd2(m) = MagicWd2;
	mCur += bytes;
	mTotal += bytes;
	if ( mCur > mMax ) mMax = mCur;
	return( bufstart(m) );
}

char *
mRealloc(p, n, file, line)
register char *p;
char *file;
unsigned n, line;
{
	register Trax *m;
	
	if ( p == NULL )
	{
		(*mError)("realloc: NULL ptr; attempted at %s(%d)\n", file?file:"???", line);
		return( NULL );
	}
	m = (Trax *) realloc(mInfo(p), n+sizeof(Trax)+(2*sizeof(char)));
	if ( m == NULL )
	{
		(*mError)("realloc: can't reallocate to %d bytes at %s(%d)\n", n, file?file:"???", line);
		return( NULL );
	}
	mCur += (n - m->size);
	if ( n > m->size ) mTotal += (n - m->size);
	if ( mCur > mMax ) mMax = mCur;
	m->size = n;
	magicwd1(m) = MagicWd1;
	magicwd2(m) = MagicWd2;
	if ( mInfo(p) == m ) return( p );
	/*
	 * Buffer has been moved.  Must adjust 'prev' pointer of next buf in list
	 * also, adjust 'next' pointer of prev buf in list
	 */
	if ( m->next != NULL ) (m->next)->prev = m;
	if ( m->prev != NULL ) (m->prev)->next = m;
    else group[m->code] = m;
	return( bufstart(m) );
}

mFree(p, file, line)
char *p;
{
	register Trax *m;
	
	if ( p == NULL )
	{
		(*mError)("free: NULL ptr; attempted at %s(%d)\n", file?file:"???", line);
		return;
	}
	if ( odd(p) )
	{
		(*mError)("free: invalid ptr (odd address); attempted at %s(%d)\n", file?file:"???", line);
		return;
	}
	m = mInfo(p);
	if ( m->mword != ValidBuf )
	{
		(*mError)("free: invalid ptr (unallocated); attempted at %s(%d)\n", file?file:"???", line);
		return;
	}
	if ( badmagic(m) )
	{
		(*mError)("free: memory corruption; code %d attempted at %s(%d) allocated at %s(%d)\n",
					m->code,
					file?file:"???",
					line,
					m->file?m->file:"???",
					m->line);
		return;
	}
	if ( m->prev != NULL ) (m->prev)->next = m->next;
	else group[m->code] = m->next;
	if ( m->next != NULL ) (m->next)->prev = m->prev;
	m->mword = 0;
	mCur -= m->size;
	free(m);
}

mMark(code)
unsigned code;
{
	if ( code >= nCodes )
	{
		(*mError)("mMark: memory code %d out of range 0..%d\n", code, nCodes-1); 
		return;
	}
	Code = code;
}

mWeed(code)
unsigned code;
{
	register Trax *m, *p;
	
	if ( code >= nCodes )
	{
		(*mError)("mWeed: memory code %d out of range 0..%d\n", code, nCodes-1); 
		return;
	}
	if ( code == 0 ) code = Code;
	m = group[code];
	while ( m != NULL )
	{
		p = m->next;
		if ( badmagic(m) )
		{
			(*mError)("mWeed: memory corruption; code %d allocated at %s(%d)\n",
						code,
						m->file?m->file:"???",
						m->line);
		}
		else
		{
			m->mword = 0;
			mCur -= m->size;
			free( m );
		}
		m = p;
	}
	group[code] = NULL;
}

mWipe()
{
	int i;
	for (i=0; i<nCodes; i++) mWeed(i);
}

mHist(f)
FILE *f;
{
	int i;
	for (i=0; i<nCodes; i++)
	{
		if ( group[i] != NULL ) mLog(f, i);
	}
}

mLog(f, code)
FILE *f;
unsigned code;
{
	register Trax *p;
	
	if ( code >= nCodes )
	{
		(*mError)("mWeed: memory code %d out of range 0..%d\n", code, nCodes-1); 
		return;
	}
	p = group[code];
	fprintf(f, "Memory Log of memory code %d:\n\n", code);
	while ( p != NULL )
	{
		fprintf(f, "[%s(%d) size %d]%s\n",
					p->file, p->line, p->size,
					badmagic(p)?" (corrupt)":"");
		p = p->next;
	}
	fprintf(f, "\n");
}

void
defErr(err, a1, a2, a3, a4, a5, a6)
char *err;
long a1, a2, a3, a4, a5, a6;
{
	fprintf(stderr, err, a1, a2, a3, a4, a5, a6);
}
