/*
 * File: pascal.g
 *
 * P a s c a l  G r a m m a r
 *
 * Some of the grammar could be more readable if it were spread out.  However,
 * to show the concise nature of ANTLR, I am squishing it (show off).
 * Theoretically, it would be possible to put the entire grammar into one
 * rule with lots-o-parenthesis and duplicate subrules.
 *
 * This grammar is considered Public Domain and is distributed as is.
 *
 * Terence Parr (c) 1990
 * Purdue University
 * January 1990
 *
 */
#attrib <<
			#include "sym.h"
			#define aStackSize	400
			typedef Sym *Attrib;
			#define aCreate(attr)	{*(attr) = CurSym;}
			extern Sym *CurSym;
		>>

<<
Sym *CurSym;
static int level = 0;
>>

#token TypeID
#token VarID
#token ProcID
#token ProgID
#token FuncID
#token ConstID
#token "[\t\ ]"		<< LexSkip(); >>
#token "\n"			<< lex_line++; LexSkip(); >>
#token "\{~[\}]*\}"	<< LexSkip(); >>


/* G r o s s  S y n t a x  C o n t r o l */


program		:	<<Sym *globals=NULL, *p;>>
				"program" WORD		<<$2->token=ProgID;aSymAdd($2->symbol,$2);>>
				<<printf("program %s level %d:\n\n", $2->symbol, level);>>
			 	<<aSymScope(&globals);>>
				"\(" idList "\)"	<<addlist($4,VarID,level);>>
				";"
				block "."
				<<
				  p = aSymRmScope(&globals);
				  printf("global (level %d) symbols:", level);
				  for (; p != NULL; p=p->scope) printf(" %s", p->symbol);
				  printf("\n\n");
				>>
				"@"
			;

block		:	<<Sym **oldscope = aSymScope(NULL);>>
				decl
				( routine <<aSymScope(oldscope);>> )*
				"begin"
					slist
				"end"
			;

routine		:	<<int tok; Sym *rout,*p,*locals=NULL;>>
				("procedure" <<tok=ProcID;>> | "function" <<tok=FuncID;>>)
				(	WORD <<rout=$1; $1->token=tok; aSymAdd($1->symbol,$1);>>
				|	ProcID
				|	FuncID
				)
				<<aSymScope(&locals); level++;>>
				{"\(" parmGroup (";" parmGroup)* "\)"}
				{ ":" TypeID }
				";"
				block ";"
				<<
				  p = aSymRmScope(&locals);
				  printf("subprogram %s level %d:\n", rout->symbol, --level);
				  printf("local (level %d) symbols:", level+1);
				  for (; p != NULL; p=p->scope) printf(" %s", p->symbol);
				  printf("\n\n");
				>>
			;

parmGroup	:	{"var"} idList ":" TypeID <<addlist($2,VarID,level);>>
			;

slist		:	statement ( ";" statement )* ;

statement	:	{UINT ":"}
				{	(var | FuncID) ":=" expr
				|	ProcID { "\(" expr ("," expr)* "\)" }
				|	"begin" slist "end"
				|	"if" expr "then" statement {"else" statement}
				|	"case" expr "of"
						(constant ("," constant)* ":" statement ";")*
					"end"
				|	"while" expr "do" statement
				|	"repeat" slist "until" expr
				|	"for" VarID ":=" expr ("to"|"downto") expr "do" statement
				|	"with" var ( "," var )* "do" statement
				|	"goto" UINT
				}
			;


/* D e c l a r a t i o n  S e c t i o n */


decl		:	("label" UINT ( "," UINT )* ";")*
				{	"const" WORD "=" constant ";"
					<<$2->token = ConstID; aSymAdd($2->symbol,$2);>>
					(	WORD <<$1->token = ConstID; aSymAdd($1->symbol,$1);>>
						"=" constant ";"
					)*
				}
				{	"type" WORD	"=" type ";"
					<<$2->token = TypeID; aSymAdd($2->symbol,$2);>>
					(	WORD	<<$1->token = TypeID; aSymAdd($1->symbol,$1);>>
						"=" type ";"
					)*
				}
				{	"var" idList ":" type ";"	<<addlist($2,VarID,level);>>
					( idList ":" type ";"		<<addlist($1,VarID,level);>>
					)*
				}
			;

simpleType	:	TypeID
			|	"\(" wordList <<addlist($2,ConstID,level);>> "\)"
			|	constant ".." constant
			;

type		:	simpleType
			|	"^" TypeID
			|	{PACKED} structType
			;

structType	:	"array" "\[" simpleType ("," simpleType)* "\]" "of" type
			|	"file" "of" type
			|	"set" "of" simpleType
			|	"record" fixedPart {";"} (fixedPart {";"} )* {variantPart} "end"
			;

fixedPart	:	idList ":" type ;

variantPart	:	"case"
				{WORD ":"}
				TypeID "of" variant {";"} (variant {";"})*
			;

variant		:	constant ( "," constant )*
				":" "\(" fixedPart {";"} (fixedPart {";"})* {variantPart} "\)"
			;

/* Use 'next' field of symbol rec to link elements together into idlist */
idList		:	<<Sym *list=NULL;>>			/* Return list of id's */
				(	WORD	<<$1->next=list; list=$1;>>
				|	VarID	<<{Sym *p=aSymNew($1->symbol);
							   p->next=list; list=p;}>>
				)
				( ","
					(	WORD <<$1->next=list; list=$1;>>
					|	VarID <<{Sym *p=aSymNew($1->symbol);
								 p->next=list; list=p;}>>
					)
				)*
				<<$0 = list;>>
			;

wordList	:	<<Sym *list=NULL;>>			/* Return list of words */
				WORD	<<$1->next=list; list=$1;>>
				( "," WORD <<$2->next=list; list=$2;>> )*
				<<$0 = list;>>
			;

constant	:	{"\+"|"\-"} (ConstID | UINT | UNUM)
			|	LITERAL
			;

uconstant	:	ConstID
			|	UNUM
			|	UINT
			|	"nil"
			|	LITERAL
			;
			

/* E x p r e s s i o n  S e c t i o n */


expr		:	simpleExpr {("="|"\<"|"\>"|"\<\>"|"\<="|"\>="|"in") simpleExpr} ;

simpleExpr	:	{"\+"|"\-"} term {("\+"|"\-"|"or") term} ;

term		:	factor {("\*"|"/"|"div"|"mod"|"and") factor} ;

factor		:	uconstant
			|	var
			|	FuncID { "\(" expr ("," expr)* "\)" }
			|	"\(" expr "\)"
			|	"not" factor
			|	"\[" {expr {".." expr}} ("," expr {".." expr})* "\]"
			;

var			:	(VarID|WORD) ref
			;

ref			:	"\[" expr ("," expr)* "\]" ref
			|	"^" ref
			|	"." (WORD|VarID) ref
			|
			;

#token LITERAL "' ~[\0-\31']* '"
#token UINT "[0-9]+"
#token SINT	"{\+|\-} [0-9]+"
#token UNUM	"[0-9]* {. [0-9]*} {E{\+|\-} [0-9]+}"	/* Floating point */
#token WORD "[a-zA-Z] [a-zA-Z0-9]*"
		<<
		{
			register Sym *p;

			p = aSymGet(LexText);
			if ( p == NULL ) p = aSymNew(LexText, WORD);
			else Token = p->token;
			CurSym = p;
		}
		>>
